#
# (C) COPYRIGHT 2020, LabNS - Laboratório de Nano-Espectroscopia / UFMG
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, and to permit persons to whom the Software is furnished 
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

#' @file 2d_band_fitter.R
#' @author Joao Luiz Elias Campos \email{jlecampos@gmail.com}
#' @date August 01, 2020

library(DBI)
library(progress)
source('R/colormap.R')
source('R/crop_data.R')
source('R/baseline.R')
source('R/mimage.R')
source('R/lorentzian.R')
source('R/save_hspe.R')

# Fit function with four lorentzian
fitter = function(x,y0,A1,xc1,w1,A2,xc2,w2,A3,xc3,w3,A4,xc4,w4)
{
  .expr1 = pi*(4*(x-xc1)^2+w1^2)
  .expr2 = pi*(4*(x-xc2)^2+w2^2)
  .expr3 = pi*(4*(x-xc3)^2+w3^2)
  .expr4 = pi*(4*(x-xc4)^2+w4^2)
  .value1 = 2*A1*w1/.expr1
  .value2 = 2*A2*w2/.expr2
  .value3 = 2*A3*w3/.expr3
  .value4 = 2*A4*w4/.expr4
  .value  = y0+.value1+.value2+.value3+.value4
  .args = as.list(match.call()[c("y0","A1","xc1","w1","A2","xc2","w2",
                                      "A3","xc3","w3","A4","xc4","w4")])
  if(all(unlist(lapply(.args,is.name)))) {
    .grad = array(0,c(length(x),13),list(NULL,c("y0",
                                                "A1","xc1","w1",
                                                "A2","xc2","w2",
                                                "A3","xc3","w3",
                                                "A4","xc4","w4")))
    .grad[,"y0"] = 1
    .grad[,"A1"] = 2*w1/.expr1
    .grad[,"xc1"] = 16*A1*w1*pi*(x-xc1)/.expr1^2
    .grad[,"w1"] = 2*A1/.expr1-4*A1*pi*w1^2/.expr1^2
    .grad[,"A2"] = 2*w2/.expr2
    .grad[,"xc2"] = 16*A2*w2*pi*(x-xc2)/.expr2^2
    .grad[,"w2"] = 2*A2/.expr2-4*A2*pi*w2^2/.expr2^2
    .grad[,"A3"] = 2*w3/.expr3
    .grad[,"xc3"] = 16*A3*w3*pi*(x-xc3)/.expr3^2
    .grad[,"w3"] = 2*A3/.expr3-4*A3*pi*w3^2/.expr3^2
    .grad[,"A4"] = 2*w4/.expr4
    .grad[,"xc4"] = 16*A4*w4*pi*(x-xc4)/.expr4^2
    .grad[,"w4"] = 2*A4/.expr4-4*A4*pi*w4^2/.expr4^2
    attr(.value,"gradient") = .grad
  }
  return(.value)
}

args = commandArgs(trailingOnly=TRUE)

if(length(args) != 1) {
  cat('usage: Rscript 2d_band_fitter.R <filename>\n')
  q()
} else if(!file.exists(args[1])) {
  cat(paste0("File ",args[1]," does not exist\n"))
  q()
}

cat('Loading input file...')
con = dbConnect(RSQLite::SQLite(),args[1])
hdr = dbGetQuery(con,'SELECT * FROM HEADER')
spc = dbGetQuery(con,'SELECT * FROM SPECTRUM')
dbDisconnect(con)
cat('\n')

cat('Loading colormap...')
cmap = colormap('fire')
cat('\n')

fname = strsplit(args[1],'[.]')[[1]][1]
pdf(file=paste0(fname,"_2d_fit.pdf"))

# Read raman shift values
xs = readBin(hdr$frequency[[1]],what='double',size=4,n=hdr$size)

# Initial values for fit
vstr = list(y0=-35,A1=10428,xc1=2586,w1=30,A2=41098,xc2=2631,w2=27,
                   A3=28458,xc3=2651,w3=28,A4=11391,xc4=2669,w4=21)
crtl = list(maxiter=100000,minFactor=1e-10,warnOnly=TRUE)

dx = c(0,diff(xs))

# Fit all spectrum
p = matrix(nrow=nrow(spc),ncol=13)
h = rep(0,nrow(spc))
pbar = txtProgressBar(min=1,max=nrow(spc),style=3)
for(i in 1:nrow(spc)) {
# read spectrum
  ys = readBin(spc$value[i][[1]],what= 'double',size=4,n=hdr$size)

# fit lorentzians
  f = nls(y~fitter(x,y0,A1,xc1,w1,A2,xc2,w2,A3,xc3,w3,A4,xc4,w4),
          data=data.frame(x=xs,y=ys),start=vstr,control=crtl)

# compute fitted spectrum
  y = predict(f,newdata=data.frame(x=xs))

# compute fitted curve area
  A = sum(dx*y)

# compute equivalent FWHM
  w[i] = 2*A/(pi*max(y))

# store fit parameters
  p[i,] = coef(f)
  setTxtProgressBar(pbar,i)
}

# Display equivalent FWHM
v = matrix(w,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Write data to a file
w = cbind(p,w)
colnames(w) = c('y0','A1','xc1','w1','A2','xc2','w2','A3','xc3','w3',
                'A4','xc4','w4','wm')
write.table(w,file='2d_band_bl_04_fparam.txt',row.names=F)

# End of File

