#
# (C) COPYRIGHT 2020, LabNS - Laboratório de Nano-Espectroscopia / UFMG
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, and to permit persons to whom the Software is furnished 
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

#' @file g_band_fitter.R
#' @author Joao Luiz Elias Campos \email{jlecampos@gmail.com}
#' @date August 01, 2020

library(DBI)
library(progress)
source('R/colormap.R')
source('R/crop_data.R')
source('R/baseline.R')
source('R/mimage.R')
source('R/lorentzian.R')
source('R/save_hspe.R')

# Fit function to be used:
# three lorentzians
fitter = function(x,y0,A1,xc1,w1,A2,xc2,w2,A3,xc3,w3)
{
  .expr1 = pi*(4*(x-xc1)^2+w1^2)
  .expr2 = pi*(4*(x-xc2)^2+w2^2)
  .expr3 = pi*(4*(x-xc3)^2+w3^2)
  .value1 = 2*A1*w1/.expr1
  .value2 = 2*A2*w2/.expr2
  .value3 = 2*A3*w3/.expr3
  .value  = y0+.value1+.value2+.value3
  .args = as.list(match.call()[c("y0","A1","xc1","w1","A2","xc2","w2",
                                      "A3","xc3","w3")])
  if(all(unlist(lapply(.args,is.name)))) {
    .grad = array(0,c(length(x),10),list(NULL,c("y0",
                                                "A1","xc1","w1",
                                                "A2","xc2","w2",
                                                "A3","xc3","w3")))
    .grad[,"y0"] = 1
    .grad[,"A1"] = 2*w1/.expr1
    .grad[,"xc1"] = 16*A1*w1*pi*(x-xc1)/.expr1^2
    .grad[,"w1"] = 2*A1/.expr1-4*A1*pi*w1^2/.expr1^2
    .grad[,"A2"] = 2*w2/.expr2
    .grad[,"xc2"] = 16*A2*w2*pi*(x-xc2)/.expr2^2
    .grad[,"w2"] = 2*A2/.expr2-4*A2*pi*w2^2/.expr2^2
    .grad[,"A3"] = 2*w3/.expr3
    .grad[,"xc3"] = 16*A3*w3*pi*(x-xc3)/.expr3^2
    .grad[,"w3"] = 2*A3/.expr3-4*A3*pi*w3^2/.expr3^2
    attr(.value,"gradient") = .grad
  }
  return(.value)
}

# Input file from command line
args = commandArgs(trailingOnly=TRUE)

if(length(args) != 1) {
  cat('usage: Rscript g_band_fitter.R <filename>\n')
  q()
} else if(!file.exists(args[1])) {
  cat(paste0("File ",args[1]," does not exist\n"))
  q()
}

# Loading input data from a HSPE file
# The HSPE file is a sqlite file.
cat('Loading input file...')
con = dbConnect(RSQLite::SQLite(),args[1])
hdr = dbGetQuery(con,'SELECT * FROM HEADER')
spc = dbGetQuery(con,'SELECT * FROM SPECTRUM')
dbDisconnect(con)
cat('\n')

cat('Loading colormap...')
cmap = colormap('fire')
cat('\n')

fname = strsplit(args[1],'[.]')[[1]][1]
pdf(file=paste0(fname,"_g_fit.pdf"))

# Get raman shift values from file HEADER
xs = readBin(hdr$frequency[[1]],what='double',size=4,n=hdr$size)

# Compute initial values for the nolinear fit function
dx = mean(diff(xs)) * 3
vstr = list(y0=1,A1=10428,xc1=1533,w1=27,A2=41098,xc2=1582,w2=27,
                   A3=28458,xc3=1614,w3=28)
crtl = list(maxiter=100000000,minFactor=1e-10,warnOnly=TRUE)

# Define the value ranges for each variable to be fitted.
lower = c(0,0,1410,0,0,1582-dx,0,0,1582,0)
upper = c(300,10^30,1582,10^30,
              10^30,1582+dx,10^30,
              10^30,1750,10^30)

tv = matrix(nrow=nrow(spc),ncol=hdr$size,byrow=TRUE)
h = matrix(nrow=nrow(spc),ncol=3,byrow=TRUE)
p = matrix(nrow=nrow(spc),ncol=10)
colnames(p) = c('y0','A1','xc1','w1','A2','xc2','w2','A3','xc3','w3')
iscv = rep(FALSE,nrow(spc))
pbar = txtProgressBar(min=1,max=nrow(spc),style=3)
for(i in 1:nrow(spc)) {
# read intensity values from input file.
  ys = readBin(spc$value[i][[1]],what= 'double',size=4,n=hdr$size)

# fit process
  f = nls(y~fitter(x,y0,A1,xc1,w1,A2,xc2,w2,A3,xc3,w3),
          data=data.frame(x=xs,y=ys),
          algorithm='port',
          start=vstr,
          lower=lower,
          upper=upper,
          control=crtl)
# fitted spectrum
  tv[i,] = predict(f,newdata=data.frame(x=xs))

# fit parameters
  p[i,] = coef(f)

# convergency status for each spectrum
  iscv[i] = f$convInfo$isConv

# spectrum intensity for each fitted band.
  h[i,1] = 2*p[i,2]/pi/p[i,4]
  h[i,2] = 2*p[i,5]/pi/p[i,7]
  h[i,3] = 2*p[i,8]/pi/p[i,10]
  if(!is.finite(h[i,1]) || 
      p[i,3] > 1533+dx || p[i,3] < 1533-dx || p[i,4] < 1) {
    h[i,1] = 0
  }
  if(!is.finite(h[i,3]) || 
      p[i,9] > 1614+dx || p[i,9] < 1614-dx || p[i,10] < 1) {
    h[i,3] = 0
  }

# display each fitted spectrum
  plot(xs,ys,cex=0.5,pch=15,main=paste(i,iscv[i]))
  lines(xs,tv[i,],col='orange')
  lines(xs,lorentzian(xs,p[i,1],p[i,2],p[i,3],p[i,4]),col='red')
  lines(c(p[i,3],p[i,3]),c(p[i,1],p[i,1]+h[i,1]),col='red')
  lines(xs,lorentzian(xs,p[i,1],p[i,5],p[i,6],p[i,7]),col='darkgreen')
  lines(c(p[i,6],p[i,6]),c(p[i,1],p[i,1]+h[i,2]),col='darkgreen')
  lines(xs,lorentzian(xs,p[i,1],p[i,8],p[i,9],p[i,10]),col='blue')
  lines(c(p[i,9],p[i,9]),c(p[i,1],p[i,1]+h[i,3]),col='blue')
  setTxtProgressBar(pbar,i)
}

# Save fitted spectrum in a HSPE file
save_hspe(name=paste0(fname,"_fit.hspe"),x=xs,y=tv,width=hdr$width,
          height=hdr$height)

# Write fit parameter to a file.
write.table(p,file='g_band_04_bl_fparam',row.names=F)

# Remove intensity values saturation for the first band and display it
q = quantile(h[which(h[,1] > 0),1],prob=c(0.01,0.99))
v = h[,1]
v[v < q[1]] = q[1]
v[v > q[2]] = q[2]
v = matrix(v,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Display G FWHM map
v = matrix(p[,7],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Remove intensity values saturation for the third band and display it
q = quantile(h[which(h[,3] > 0),3],prob=c(0.01,0.99))
v = h[,3]
v[v < q[1]] = q[1]
v[v > q[2]] = q[2]
v = matrix(v,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# End of File

