#
# (C) COPYRIGHT 2020, LabNS - Laboratório de Nano-Espectroscopia / UFMG
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, and to permit persons to whom the Software is furnished 
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

#' @file pca.R
#' @author Joao Luiz Elias Campos \email{jlecampos@gmail.com}
#' @date July 25, 2020

library(DBI)
library(progress)
library(pracma)
library(factoextra)
source('R/crop_data.R')
source('R/baseline.R')
source('R/colormap.R')
source('R/mimage.R')
source('R/reconstruct.R')

# Fit function for g_band_bl_tp file
fitter1 = function(x,a,b,e,d) {
  .value = a*e + b*d
  .args = as.list(match.call()[c("a","b")])
  if(all(unlist(lapply(.args,is.name)))) {
    .grad = array(0,c(length(.value),2),list(NULL,c("a","b")))
    .grad[,"a"] = e
    .grad[,"b"] = d
    attr(.value,"gradient") = .grad
  }
  return(.value)
}

# Fit function for g_band_bl_04 file
fitter2 = function(x,a,b,c,e,d,v) {
  .value = a*e + b*d + c*v
  .args = as.list(match.call()[c("a","b","c")])
  if(all(unlist(lapply(.args,is.name)))) {
    .grad = array(0,c(length(.value),3),list(NULL,c("a","b","c")))
    .grad[,"a"] = e
    .grad[,"b"] = d
    .grad[,"c"] = v
    attr(.value,"gradient") = .grad
  }
  return(.value)
}

args = commandArgs(trailingOnly=TRUE)

if(length(args) != 1) {
  cat('usage: Rscript pca.R <filename>\n')
  q()
} else if(!file.exists(args[1])) {
  cat(paste0("File ",args[1]," does not exist\n"))
  q()
}

fname = strsplit(args[1],'[.]')[[1]][1]
pdf(file=paste0(fname,"_pca.pdf"))

export = TRUE

cat('Loading colormap...')
cmap = colormap('fire')
cat('\n')

cat('Loading input file...')
con = dbConnect(RSQLite::SQLite(),args[1])
hdr = dbGetQuery(con,'SELECT * FROM HEADER')
spc = dbGetQuery(con,'SELECT * FROM SPECTRUM')
dbDisconnect(con)
cat('\n')

# Read raman shift value and crop at G band range
xs = readBin(hdr$frequency[[1]],what='double',size=4,n=hdr$size)
xid = crop_data(xs,vmin=1450,vmax=1680,vindex=TRUE)
xs = xs[xid]
print(range(xs))

cat('Building samples:\n')
pbar = txtProgressBar(min=1,max=nrow(spc),style=3)
tv = matrix(nrow=nrow(spc),ncol=length(xs),byrow=TRUE)
for(i in 1:nrow(spc)) {
# read spectrum value
  ys = readBin(spc$value[i][[1]],what= 'double',size=4,n=hdr$size)

# crop at G band range
  tv[i,] = ys[xid]
  setTxtProgressBar(pbar,i)
}
cat('\n')

cat('Doing PCA...')
pca = prcomp(as.data.frame(tv),center=TRUE,scale=FALSE)
cat('\n')

screeplot(pca,xlab='PCs',main='Proportion of Variance')

plot(pca$x[,1],pca$x[,2],cex=0.5,pch=15,xlab='PC1',ylab='PC2')
text(pca$x[,1],pca$x[,2],1:nrow(spc),cex=0.6,pos=4,of=0.2)
abline(v=-100,col='red')

# Display PC1 map
v = matrix(pca$x[,1],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Display PC2 map
v = matrix(pca$x[,2],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Display PC3 map
v = matrix(pca$x[,3],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Display PC4 map
v = matrix(pca$x[,4],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
mimage(v,col=cmap,scale=TRUE)

# Reconstruct data from PC1 only, PC2 only and PC1+PC2
# plot all tree spectrum
m1 = reconstruct(pca,ids=c(1))
m2 = reconstruct(pca,ids=c(2))
m3 = reconstruct(pca,ids=c(1,2))
for(i in 1:nrow(m1)) {
  plot(xs,m1[i,]/max(m1[i,]),type='l',main=i)
  lines(xs,m2[i,]/max(m2[i,]),col='red')
  lines(xs,m3[i,]/max(m3[i,]),col='blue')
}

fname = strsplit(args[1],'[.]')[[1]][1]

# Export a recontructed sample spectrum
if(export) {
  write.table(cbind(xs,m1[101,]/max(m1[101,]),m1[197,]/max(m1[197,])),
              file=paste0(fname,"_pc1.txt"),col.names=FALSE, row.names=FALSE)

  v = matrix(pca$x[,1],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  write.table(v,file=paste0(fname,"_pc1_map.txt"),col.names=FALSE,
              row.names=FALSE)
}

# Process TP file
if(fname == 'g_band_bl_tp') {
# load reference spectrum
  edg = read.table('g_band_bl_tp-edg.dat')
  dmn = read.table('g_band_bl_tp-dmn.dat')

# resample reference spectrum
  fe = approx(edg[,1],edg[,2]-min(edg[,2]),xs,rule=2)
  fd = approx(dmn[,1],dmn[,2]-min(dmn[,2]),xs,rule=2)
  ce = rep(NA,nrow(tv))
  cd = rep(NA,nrow(tv))

# compute spectrum correlation to the reference spacetrum
  for(i in 1:nrow(tv)) {
    ce[i] = cor(fe$y/max(fe$y),tv[i,]/max(tv[i,]))
    cd[i] = cor(fd$y/max(fd$y),tv[i,]/max(tv[i,]))
  }

# remove pixel saturation
  q = quantile(ce,probs=c(0.05,0.95))
  ce[ce < q[1]] = q[1]
  ce[ce > q[2]] = q[2]

  q = quantile(cd,probs=c(0.05,0.95))
  cd[cd < q[1]] = q[1]
  cd[cd > q[2]] = q[2]

# display edge correlation map
  v = matrix(ce,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,scale=TRUE,col=cmap)

  if(export) {
    write.table(v,file=paste0(fname,"_ce_map.txt"),col.names=FALSE,
                row.names=FALSE)
  }
 
# display domain correlation map
  v = matrix(cd,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,scale=TRUE,col=cmap)

  if(export) {
    write.table(v,file=paste0(fname,"_cd_map.txt"),col.names=FALSE,
                row.names=FALSE)
  }

# normalize reference spectrum
  e = fe$y/max(fe$y)
  d = fd$y/max(fd$y)

# compute reference spectrum contributions to each input spectrum
  cat('Process fitting:\n')
  pbar = txtProgressBar(min=1,max=nrow(tv),style=3)
  pv = matrix(nrow=nrow(tv),ncol=2,byrow=TRUE)
  mse = rep(NA,nrow(tv))
  for(i in 1:nrow(tv)) {
    y = tv[i,]/max(tv[i,])
    f = nls(y~fitter1(x,a,b,e,d),data=data.frame(x=xs,y=y),
                      start=list(a=0.5,b=0.5),algorithm='port',
                      lower=c(0,0),upper=c(1,1),#trace=TRUE,
                      control=list(maxiter=1000,minFactor=1/2000,warnOnly=TRUE))
    pv[i,] = coef(f)
    mse[i] = sum(residuals(f)^2)/length(y)
    setTxtProgressBar(pbar,i)
  }
  cat('\n')

# display edge contribution map
  v = matrix(pv[,1],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,col=cmap,scale=TRUE,main='E')
  if(export) {
    write.table(v,file=paste0(fname,"_fit_acoef.txt"),col.names=FALSE,
                row.names=FALSE)
  }

# display domain contribution map
  v = matrix(pv[,2],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,col=cmap,scale=TRUE,main='D')
  if(export) {
    write.table(v,file=paste0(fname,"_fit_bcoef.txt"),col.names=FALSE,
                row.names=FALSE)
  }
# Process 04 file
} else {
# load reference spectrum
  edg = read.table('g_band_bl_04-edg.dat')
  dmn = read.table('g_band_bl_04-dmn.dat')
  vrtx = read.table('g_band_bl_04-vrtx.dat')

# resample reference spectrum
  fe = approx(edg[,1],edg[,2]-min(edg[,2]),xs,rule=2)
  fd = approx(dmn[,1],dmn[,2]-min(edg[,2]),xs,rule=2)
  fv = approx(vrtx[,1],vrtx[,2]-min(vrtx[,2]),xs,rule=2)
  ce = rep(NA,nrow(tv))
  cd = rep(NA,nrow(tv))
  cv = rep(NA,nrow(tv))

# compute spectrum correlation to the reference spacetrum
  for(i in 1:nrow(tv)) {
    ce[i] = cor(fe$y/max(fe$y),tv[i,]/max(tv[i,]))
    cd[i] = cor(fd$y/max(fd$y),tv[i,]/max(tv[i,]))
    cv[i] = cor(fv$y/max(fv$y),tv[i,]/max(tv[i,]))
  }

# remove pixel saturation
  q = quantile(ce,probs=c(0.05,0.95))
  ce[ce < q[1]] = q[1]
  ce[ce > q[2]] = q[2]

  q = quantile(cd,probs=c(0.05,0.95))
  cd[cd < q[1]] = q[1]
  cd[cd > q[2]] = q[2]

  q = quantile(cv,probs=c(0.05,0.95))
  cv[cv < q[1]] = q[1]
  cv[cv > q[2]] = q[2]

# display edge correlation map
  v = matrix(ce,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,scale=TRUE,col=cmap)

  if(export) {
    write.table(v,file=paste0(fname,"_ce_map.txt"),col.names=FALSE,
                row.names=FALSE)
  }
 
# display domain correlation map
  v = matrix(cd,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,scale=TRUE,col=cmap)

  if(export) {
    write.table(v,file=paste0(fname,"_cd_map.txt"),col.names=FALSE,
                row.names=FALSE)
  }

# display vertex correlation map
  v = matrix(cv,nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,scale=TRUE,col=cmap)

  if(export) {
    write.table(v,file=paste0(fname,"_cv_map.txt"),col.names=FALSE,
                row.names=FALSE)
  }

# normalize reference spectrum
  e = fe$y/max(fe$y)
  d = fd$y/max(fd$y)
  v = fv$y/max(fv$y)

# compute reference spectrum contributions to each input spectrum
  cat('Process fitting:\n')
  pbar = txtProgressBar(min=1,max=nrow(tv),style=3)
  pv = matrix(nrow=nrow(tv),ncol=3,byrow=TRUE)
  mse = rep(NA,nrow(tv))
  for(i in 1:nrow(tv)) {
    y = tv[i,]/max(tv[i,])
    f = nls(y~fitter2(x,a,b,c,e,d,v),data=data.frame(x=xs,y=y),
                      start=list(a=0.3,b=0.3,c=0.3),algorithm='port',
                      lower=c(0,0,0),upper=c(1,1,1),
                      control=list(maxiter=1000,minFactor=1/2000,warnOnly=TRUE))
    pv[i,] = coef(f)
    mse[i] = sum(residuals(f)^2)/length(y)
    setTxtProgressBar(pbar,i)
  }
  cat('\n')

# display edge correlation map
  v = matrix(pv[,1],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,col=cmap,scale=TRUE,main='E')
  if(export) {
    write.table(v,file=paste0(fname,"_fit_acoef.txt"),col.names=FALSE,
                row.names=FALSE)
  }

# display domain correlation map
  v = matrix(pv[,2],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,col=cmap,scale=TRUE,main='D')
  if(export) {
    write.table(v,file=paste0(fname,"_fit_bcoef.txt"),col.names=FALSE,
                row.names=FALSE)
  }

# display vertex correlation map
  v = matrix(pv[,3],nrow=hdr$height,ncol=hdr$width,byrow=TRUE)
  mimage(v,col=cmap,scale=TRUE,main='V')
  if(export) {
    write.table(v,file=paste0(fname,"_fit_ccoef.txt"),col.names=FALSE,
                row.names=FALSE)
  }
}

# End of File

